; TIMER0 event rate = 4 times the UART baud rate
;
; R7    TX bit counter
; R6    TX shift register
; R5    RX bit counter / RX new head index / RX timeout counter
; R4    RX shift register
; R3    RX data storage pointer (last word)
; R2    RX idle counter
; R1    Temporary variables
; R0    Temporary variables (special functionality)


; Flags set for each RX byte
.define BV_UART_RX_TIMEOUT                      0x0800
.define BV_UART_RX_OVERFLOW                     0x0400
.define BV_UART_RX_BREAK                        0x0200
.define BV_UART_RX_FRAMING_ERROR                0x0100

; Alert-generating events
.define BV_UART_ALERT_RX_FIFO_ABOVE_THR         0x0001
.define BV_UART_ALERT_RX_BYTE_TIMEOUT           0x0002
.define BV_UART_ALERT_RX_BREAK_OR_FRAMING_ERROR 0x0004
.define BV_UART_ALERT_TX_FIFO_BELOW_THR         0x0008



UartEmulator:
                        ; Disable RX idle timeout
                        ld          R5, #0

                        ; Set the default number of required idle half-bits before enabling RX
                        ld          R2, [#uartTaskName/cfg/rxEnableReqIdleCount]

                        ; Initialize state machine (function pointers)
                        ld          R0, #UartEmulator/txIdle
                        st          R0, [#uartTxFptr]
                        ld          R0, #UartEmulator/rxIdle
                        st          R0, [#uartRxFptr]
                        ld          R0, #UartEmulator/sysCmd
                        st          R0, [#uartSysFptr]
                        ld          R0, #UartEmulator/rxCheckStartDisabled
                        st          R0, [#uartRxCheckStartFptr]




/mainLoop:
                        ; 0: RX (oversampling)
                        ld          R0, [#uartRxFptr]
                        wev1        #WEVSEL_TIMER0
                        jsr         R0

                        ; 1: TX
                        ld          R0, [#uartTxFptr]
                        wev1        #WEVSEL_TIMER0
                        jsr         R0

                        ; 2: RX (oversampling)
                        ld          R0, [#uartRxFptr]
                        wev1        #WEVSEL_TIMER0
                        jsr         R0

                        ; 3: System (jumps to the start of the main loop or returns)
                        ld          R0, [#uartSysFptr]
                        wev1        #WEVSEL_TIMER0
                        jmp         R0






;; TX IDLE
/txIdle:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Start transmission of a new byte?
                        ld          R6, [#uartTaskName/state/txHead]
                        ld          R1, [#uartTaskName/state/txTail]
                        cmp         R1, R6
                        beq         /txIdleNoStart
/txIdleCheckStartDelay:
                            ; Load the data byte (LSB) and delay (MSB) into the shift register
                            add         R1, #uartTaskName/input/pTxBuffer
                            ld          R6, [R1]

                            ; Proceed to "txStartBit" next
                            ld          R1, #/txStartBit
                            st          R1, [#uartTxFptr]
/txIdleNoStart:
                        ; Done
                        rts




;; TX START BIT
/txStartBit:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Delay the transmission? (remaining delay is in MSB of the shift register)
                        cmp         R6, #0x00FF
                        bleu        /txStartBitNoDelay

                            ; Decrement the delay
                            add         R6, #0xFF00
                        bra         /txStartBitDone

                        ; Otherwise ...
/txStartBitNoDelay:
                            ; Transmit the start bit
                            iobclr      #(AUXIO_UART_TX & 0x7), [#(IOP_AIODIO0_GPIODOUT + (AUXIO_UART_TX >> 3))]

                            ; Initialize the bit counter used by "txDataBit"
                            ld          R7, #8

                            ; Proceed to "txDataBit" next
                            ld          R1, #/txDataBit
                            st          R1, [#uartTxFptr]
/txStartBitDone:
                        ; Done
                        rts




;; TX DATA BIT
/txDataBit:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Transmit lsb of the shift register
                        tst         R6, #0x0001
                        bnz         /txDataBitOutput1

                        ; Transmit logical 0
                            iobclr      #(AUXIO_UART_TX & 0x7), [#(IOP_AIODIO0_GPIODOUT + (AUXIO_UART_TX >> 3))]
                        bra         /txDataBitOutputDone
/txDataBitOutput1:
                            ; Transmit logical 1
                            iobset      #(AUXIO_UART_TX & 0x7), [#(IOP_AIODIO0_GPIODOUT + (AUXIO_UART_TX >> 3))]
/txDataBitOutputDone:
                        ; Update the shift register
                        lsr         R6, #1

                        ; Proceed to "txStopBit" next when the msb has been transmitted (counter in R7
                        ; goes from 8 to 0)
                        add         R7, #-1
                        bnz         /txDataBitNotMsb
                            ld          R1, #/txStopBit
                            st          R1, [#uartTxFptr]
/txDataBitNotMsb:
                        ; Done
                        rts




;; TX STOP BIT
/txStopBit:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Proceed to "txIdle" next, for a second stop bit
                        ld          R1, #/txIdle
                        st          R1, [#uartTxFptr]

                        ; Transmit the stop bit
                        iobset      #(AUXIO_UART_TX & 0x7), [#(IOP_AIODIO0_GPIODOUT + (AUXIO_UART_TX >> 3))]

                        ; Move on the TX FIFO tail index, wrapping at UART_TX_BUFFER_SIZE
                        ld          R1, [#uartTaskName/state/txTail]
                        add         R1, #1
                        cmp         R1, #UART_TX_BUFFER_SIZE
                        bltu        /txStopBitFifoNoWrap
                            ld          R1, #0
/txStopBitFifoNoWrap:   st          R1, [#uartTaskName/state/txTail]

                        ; Done
                        rts






/rxIdle:
                        ; If the RX timeout counter is active ...
                        cmp         R5, #0
                        bleu        /rxIdleNoTimeout

                            ; Decrement it and, and if timeout now ...
                            add         R5, #-1
                            bnz         /rxIdleNoTimeout

                                ; Set the TIMEOUT flag for the last received byte
                                or          R4, #BV_UART_RX_TIMEOUT
                                st          R4, [R3]

                                ; Set the TIMEOUT event for ALERT interrupt generation
                                ld          R1, [#uartTaskName/state/alertBacklog]
                                or          R1, #BV_UART_ALERT_RX_BYTE_TIMEOUT
                                st          R1, [#uartTaskName/state/alertBacklog]

/rxIdleNoTimeout:
                        ; If RX should be enabled ...
                        ld          R0, [#uartTaskName/state/rxEnable]
                        ld          R1, [#uartTaskName/state/rxEnabled]
                        cmp         R0, R1
                        bleu        /rxIdleNoRxEnable

                            ; If the required idle period has not yet expired ...
                            cmp         R2, #0
                            beq         /rxIdleRxEnable

                                ; Decrement the countdown
                                add         R2, #-1

                                ; If the RX line is not idle now ...
                                iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                                biob1       /rxIdleRxEnableStillIdle

                                    ; Reset the countdown
                                    ld          R2, [#uartTaskName/cfg/rxEnableReqIdleCount]
/rxIdleRxEnableStillIdle:   rts

                            ; Otherwise, if the required idle period has expired ...
/rxIdleRxEnable:
                                ; Enable start-bit detection
                                ld          R1, #/rxCheckStartEnabled
                                st          R1, [#uartRxCheckStartFptr]

                                ; Indicate that RX is enabled
                                st          R0, [#uartTaskName/state/rxEnabled]
/rxIdleNoRxEnable:
                        ; Done
                        rts




/rxDisable:
                        ; Disable start-bit detection
                        ld          R0, #/rxCheckStartDisabled
                        st          R0, [#uartRxCheckStartFptr]

                        ; Do "rxIdle" next
                        ld          R0, #/rxIdle
                        st          R0, [#uartRxFptr]

                        ; Disable the RX timeout
                        ld          R5, #0

                        ; Indicate that RX is disabled
                        st          R5, [#uartTaskName/state/rxEnabled]

                        ; Reset the required number of idle periods before allowing RX to be enabled
                        ld          R2, [#uartTaskName/cfg/rxEnableReqIdleCount]

                        ; Done
                        rts




/rxStartBit:
                        ; Stop start bit detection
                        ld          R1, #/rxCheckStartDisabled
                        st          R1, [#uartRxCheckStartFptr]

                        ; Initialize the bit counter used by "rxDataBit"
                        ld          R5, #8

                        ; Clear the shift register
                        ld          R4, #0x0000

                        ; Fetch the pointer to RX head
                        ld          R3, [#uartTaskName/state/rxHead]
                        add         R3, #uartTaskName/output/pRxBuffer

                        ; Proceed to "preRxDataBit" next
                        ld          R1, #/preRxDataBit
                        st          R1, [#uartRxFptr]

                        ; Done
                        rts




/preRxDataBit:
                        ; Update the shift register
                        lsr         R4, #1

                        ; Proceed to "rxDataBit" next
                        ld          R1, #/rxDataBit
                        st          R1, [#uartRxFptr]

                        ; Done
                        rts




/rxDataBit:
                        ; Shift in a 0 or a 1 into the shift register
                        nop
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        biob0       /rxDataBitZero
                            or          R4, #0x0080
/rxDataBitZero:
                        ; If msb has been received, proceed to "preRxStopBit" next, otherwise "preRxDataBit"
                        add         R5, #-1
                        bnz         /rxDataBitNotDone
                            ld          R1, #/preRxStopBit
                        bra         /rxDataBitDone
/rxDataBitNotDone:          ld          R1, #/preRxDataBit
/rxDataBitDone:         st          R1, [#uartRxFptr]

                        ; Done
                        rts




/preRxStopBit:
                        ; Calculate the new RX FIFO head index, wrapping at UART_RX_BUFFER_SIZE
                        ld          R5, [#uartTaskName/state/rxHead]
                        add         R5, #1
                        cmp         R5, #UART_RX_BUFFER_SIZE
                        bltu        /rxStopBitFifoNoWrap
                            ld          R5, #0
/rxStopBitFifoNoWrap:
                        ; Proceed to "rxStopBit" next
                        ld          R0, #/rxStopBit
                        st          R0, [#uartRxFptr]

                        ; Done
                        rts




/rxStopBit:
                        ; If not a valid stop bit ...
                        nop
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        biob1       /rxStopBitValid

                            ; Depending on the received byte ...
                            tst         R4, #0xFF
                            bnz         /rxStopBitFramingError
/rxStopBitBreak:
                                ; If 0x00, indicate "break"
                                or          R4, #BV_UART_RX_BREAK
                            bra         /rxStopBitNotFramingError
/rxStopBitFramingError:
                                ; Otherwise, indicate framing error
                                or          R4, #BV_UART_RX_FRAMING_ERROR
/rxStopBitNotFramingError:

                            ; Set the BREAK OR FRAMING ERROR event for ALERT interrupt generation
                            ld          R1, [#uartTaskName/state/alertBacklog]
                            or          R1, #BV_UART_ALERT_RX_BREAK_OR_FRAMING_ERROR
                            st          R1, [#uartTaskName/state/alertBacklog]

                            ; Bail out and come back later
                            rts
/rxStopBitValid:
                        ; Proceed to "rxIdle" next
                        ld          R0, #/rxIdle
                        st          R0, [#uartRxFptr]

                        ; Re-enable start-bit detection
                        ld          R0, #/rxCheckStartEnabled
                        st          R0, [#uartRxCheckStartFptr]

                        ; If the updated head index equals the tail index ...
                        ld          R0, [#uartTaskName/state/rxTail]
                        cmp         R5, R0
                        bneq        /rxStopBitNotOverflow

                            ; Indicate overflow
                            or          R4, #BV_UART_RX_OVERFLOW

                            ; Store the received byte
                            st          R4, [R3]

                            ; Initialize the timeout counter, which will be used shortly in "rxIdle"
                            ld          R5, [#uartTaskName/cfg/rxByteTimeout]

                            ; Bail out without updating the head index
                            rts
/rxStopBitNotOverflow:
                        ; Store the received byte
                        st          R4, [R3]

                        ; Update the RX head index
                        st          R5, [#uartTaskName/state/rxHead]

                        ; Initialize the timeout counter, which will be used shortly in "rxIdle"
                        ld          R5, [#uartTaskName/cfg/rxByteTimeout]

                        ; Done
                        rts






/sysCmd:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

/sysCmdExit:
                        ; Exit the UART emulator?
                        ld          R1, [#uartTaskName/state/exit]
                        cmp         R1, #0
                        beq         /sysCmdNoExit
                            rts
/sysCmdNoExit:
/sysCmdRxDisable:
                        ; Disable RX?
                        ld          R0, [#uartTaskName/state/rxEnable]
                        ld          R1, [#uartTaskName/state/rxEnabled]
                        cmp         R0, R1
                        bgeu        /sysCmdNoRxDisable

                            ; Do "rxDisable" in the next upcoming RX state
                            ld          R0, #/rxDisable
                            st          R0, [#uartRxFptr]
/sysCmdNoRxDisable:
                        ; Do ALERT mask update next
                        ld          R0, #/sysAlertMaskUpdate
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop




/sysAlertMaskUpdate:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Update the ALERT mask
                        ld          R0, [#uartTaskName/cfg/alertMask]
                        st          R0, [#uartTaskName/state/alertMask]

                        ; Clear FIFO flags and currently disabled events from the backlog
                        ld          R1, [#uartTaskName/state/alertBacklog]
                        and         R1, R0
                        and         R1, #((~(BV_UART_ALERT_RX_FIFO_ABOVE_THR | BV_UART_ALERT_TX_FIFO_BELOW_THR)) & 0xF)
                        st          R1, [#uartTaskName/state/alertBacklog]

                        ; Do RX FIFO event generation next
                        ld          R0, #/sysAlertCheckRxFifo0
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop




/sysAlertCheckRxFifo0:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Calculate the number of bytes currently in the RX FIFO
                        ld          R0, [#uartTaskName/state/rxHead]
                        ld          R1, [#uartTaskName/state/rxTail]
                        sub         R0, R1
                        bpos        /sysAlertsRxCountNoWrap
                            add         R0, #UART_RX_BUFFER_SIZE
/sysAlertsRxCountNoWrap:
                        ; There is not sufficient time to compare the FIFO count with the threshold and
                        ; set flags, so we store the count temporarily in RAM
                        st          R0, [#uartRxTxFifoCount]

                        ; Continue RX FIFO event generation next
                        ld          R0, #/sysAlertCheckRxFifo1
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop




/sysAlertCheckRxFifo1:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; If the RX FIFO is at or above the specified threshold ...
                        ld          R0, [#uartRxTxFifoCount]
                        ld          R1, [#uartTaskName/cfg/alertRxFifoThr]
                        cmp         R0, R1
                        bltu        /sysAlertsRxCountNoEvent

                            ; Generate RX FIFO ABOVE THR event
                           ld          R0, [#uartTaskName/state/alertBacklog]
                           or          R0, #BV_UART_ALERT_RX_FIFO_ABOVE_THR
                           st          R0, [#uartTaskName/state/alertBacklog]
/sysAlertsRxCountNoEvent:
                        ; Do TX FIFO event generation next
                        ld          R0, #/sysAlertCheckTxFifo0
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop




/sysAlertCheckTxFifo0:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; Calculate the number of bytes currently in the TX FIFO
                        ld          R0, [#uartTaskName/state/txHead]
                        ld          R1, [#uartTaskName/state/txTail]
                        sub         R0, R1
                        bpos        /sysAlertsTxCountNoWrap
                            add         R0, #UART_TX_BUFFER_SIZE
/sysAlertsTxCountNoWrap:
                        ; There is not sufficient time to compare the FIFO count with the threshold and
                        ; set flags, so we store the count temporarily in RAM
                        st          R0, [#uartRxTxFifoCount]

                        ; Continue TX FIFO event generation next
                        ld          R0, #/sysAlertCheckTxFifo1
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop




/sysAlertCheckTxFifo1:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; If the TX FIFO is at or below the specified threshold ...
                        ld          R0, [#uartRxTxFifoCount]
                        ld          R1, [#uartTaskName/cfg/alertTxFifoThr]
                        cmp         R0, R1
                        bgtu        /sysAlertsTxCountNoEvent

                            ; Generate TX FIFO BELOW THR event
                            ld          R0, [#uartTaskName/state/alertBacklog]
                            or          R0, #BV_UART_ALERT_TX_FIFO_BELOW_THR
                            st          R0, [#uartTaskName/state/alertBacklog]
/sysAlertsTxCountNoEvent:
                        ; Do ALERT interrupt generation next
                        ld          R0, #/sysAlertGen
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop




/sysAlertGen:
                        ; Check for start of RX
                        ld          R0, [#uartRxCheckStartFptr]
                        iobtst      #(AUXIO_UART_RX & 0x7), [#(IOP_AIODIO0_GPIODIN + (AUXIO_UART_RX >> 3))]
                        jsr         R0

                        ; If the previous events have been handled by the application ...
                        ld          R0, [#uartTaskName/state/alertEvents]
                        cmp         R0, #0x0000
                        bneq        /sysAlertsNoGen

                            ; Mask out the disabled events from the backlog
                            ld          R0, [#uartTaskName/state/alertBacklog]
                            ld          R1, [#uartTaskName/state/alertMask]
                            and         R0, R1

                            ; If there are any enabled events in the backlog ...
                            bz          /sysAlertsNoGen

                                ; Generate alert interrupt
                                st          R0, [#uartTaskName/state/alertEvents]
                                iobset      #IOB_EVCTL_SWEV_ALERT, [#IOP_EVCTL_SWEVSET]

                                ; Clear the backlog
                                ld          R0, #0x0000
                                st          R0, [#uartTaskName/state/alertBacklog]

                                ; Do system commands next
                                ld          R1, #/sysCmd
                                st          R1, [#uartSysFptr]

                                ; Keep on executing from the start of the main loop
                                jmp         /mainLoop
/sysAlertsNoGen:
                        ; Do system commands next
                        ld          R0, #/sysCmd
                        st          R0, [#uartSysFptr]

                        ; Keep on executing from the start of the main loop
                        jmp         /mainLoop






/rxCheckStartDisabled:
                        ; RX is already active, or checking is disabled
                        bra         /rxCheckStartIdle




/rxCheckStartEnabled:
                        ; RX is enabled and currently idle, so check for start bit
                        biob1       /rxCheckStartIdle
                            ld          R0, #/rxStartBit
                            st          R0, [#uartRxFptr]
                        rts
/rxCheckStartIdle:          ld          R0, #/rxStartBit
                            nop
                        rts





uartRxTxFifoCount:      dw          #0

uartTxFptr:             dw          #0
uartRxFptr:             dw          #0
uartSysFptr:            dw          #0
uartRxCheckStartFptr:   dw          #0